# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import shutil
import tempfile
import urllib.request as req
from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast, Dict, List, Tuple
from zipfile import ZipFile
from gluonts.dataset.common import MetaData
from gluonts.dataset.repository._tsf_datasets import Dataset as MonashDataset
from gluonts.dataset.repository._tsf_datasets import (
    save_datasets,
    save_metadata,
)
from gluonts.dataset.repository._tsf_reader import TSFReader
from gluonts.dataset.repository.datasets import materialize_dataset
from tsbench.config.dataset.preprocessing.filters import (
    AbsoluteValueFilter,
    ConstantTargetFilter,
    EndOfSeriesCutFilter,
    MinLengthFilter,
)
from ._base import DatasetConfig
from .preprocessing import Filter, read_transform_write


@dataclass(frozen=True)
class GluonTsDatasetConfig(DatasetConfig):  # pylint: disable=abstract-method
    """
    A dataset configuration for datasets obtained directly via GluonTS.
    """

    def generate(self) -> None:
        if self.root.exists():
            return

        (self.root / "gluonts").mkdir(parents=True)

        # Download data and move to our own managed directory
        with tempfile.TemporaryDirectory() as directory:
            self._materialize(Path(directory))
            source = Path(directory) / self._gluonts_name

            # Copy and read metadata
            meta_file = self.root / "gluonts" / "metadata.json"
            shutil.copyfile(source / "metadata.json", meta_file)
            meta = MetaData.parse_file(meta_file)

            # Copy the data and apply filters
            filters = self._filters(
                self._prediction_length_multiplier
                * cast(int, meta.prediction_length)
            )
            read_transform_write(
                self.root / "gluonts" / "train" / "data.json",
                filters=filters
                + [EndOfSeriesCutFilter(cast(int, meta.prediction_length))],
                source=source / "train" / "data.json",
            )
            read_transform_write(
                self.root / "gluonts" / "val" / "data.json",
                filters=filters,
                source=source / "train" / "data.json",
            )

            # Although we increase the prediction length for the filters here, this does not
            # exclude any more data! The time series is only longer by the prediction length...
            read_transform_write(
                self.root / "gluonts" / "test" / "data.json",
                filters=self._filters(
                    (self._prediction_length_multiplier + 1)
                    * cast(int, meta.prediction_length)
                ),
                source=source / "test" / "data.json",
            )

    @property
    def _gluonts_name(self) -> str:
        return self.name()

    @property
    def _prediction_length_multiplier(self) -> int:
        # This is a legacy field and should NOT be overridden. It was accidentally set for the
        # preprocessing of some datasets though.
        return 0

    def _filters(self, prediction_length: int) -> List[Filter]:
        return [
            ConstantTargetFilter(prediction_length),
            AbsoluteValueFilter(1e18),
        ]

    def _materialize(self, directory: Path, regenerate: bool = False) -> None:
        materialize_dataset(
            self._gluonts_name, directory, regenerate=regenerate
        )


@dataclass(frozen=True)
class MonashDatasetConfig(GluonTsDatasetConfig):
    """
    A dataset configuration for datasets obtained through forecastingdata.org,
    the Monash Forecasting Repository.
    """

    @property
    def _prediction_length(self) -> int:
        raise NotImplementedError

    @property
    def _file(self) -> str:
        raise NotImplementedError

    @property
    def _record(self) -> str:
        raise NotImplementedError

    def _materialize(self, directory: Path, regenerate: bool = False) -> None:
        dataset = MonashDataset(self._file, self._record)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)

            with ZipFile(dataset.download(temp_path)) as archive:
                archive.extractall(path=temp_path)

            # only one file is exptected
            reader = TSFReader(temp_path / archive.namelist()[0])
            meta, data = reader.read()

        path = directory / self._gluonts_name
        path.mkdir()

        # Save metadata and dataset (filling in missing timestamps)
        save_metadata(
            path,
            len(data),
            _get_frequency(meta.frequency),
            self._prediction_length,
        )

        data = [
            {**d, "start_timestamp": d.get("start_timestamp", "1970-01-01")}
            for d in data
        ]
        save_datasets(path, data, self._prediction_length)


@dataclass(frozen=True)
class M3DatasetConfig(GluonTsDatasetConfig):  # pylint: disable=abstract-method
    """
    A dataset configuration shared by all M3 datasets.
    """

    def generate(self) -> None:
        if self.root.exists():
            return

        # Download the .xls file
        target = Path.home() / ".mxnet" / "gluon-ts" / "datasets" / "M3C.xls"
        if not target.exists():
            target.parent.mkdir(parents=True, exist_ok=True)
            opener = req.build_opener()
            opener.addheaders = [("User-Agent", "")]
            req.install_opener(opener)
            req.urlretrieve(
                "https://forecasters.org/data/m3comp/M3C.xls", target
            )

        # Run the generation
        super().generate()


@dataclass(frozen=True)
class KaggleDatasetConfig(GluonTsDatasetConfig):
    """
    A dataset configuration for datasets obtained directly from Kaggle.
    """

    def generate(self) -> None:
        # Check for the existence of the data
        if self.root.exists():
            return

        data_root = (
            Path.home() / ".mxnet" / "gluon-ts" / "datasets" / self.name()
        )
        if not data_root.exists():
            raise ValueError(
                f"download the dataset from Kaggle ({self._link}) and unzip it"
                f" into {data_root}"
            )

        # Extract the data and apply filters. The min length filter uses +3 as otherwise, catch22
        # features cannot be computed
        metadata, series = self._extract_data(data_root)
        filters = self._filters(metadata["prediction_length"]) + [
            MinLengthFilter(2 * metadata["prediction_length"] + 3)
        ]
        for f in filters:
            series = f(series)

        # Write everything to file
        test_dir = self.root / "gluonts" / "test"
        test_dir.mkdir(parents=True)

        with (self.root / "gluonts" / "metadata.json").open("w+") as f:
            json.dump(metadata, f)

        with (test_dir / "data.json").open("w+") as f:
            for item in series:
                json.dump(item, f)
                f.write("\n")

        # Create training and validation data
        read_transform_write(
            self.root / "gluonts" / "val" / "data.json",
            filters=[EndOfSeriesCutFilter(metadata["prediction_length"])],
            source=test_dir / "data.json",
        )
        read_transform_write(
            self.root / "gluonts" / "train" / "data.json",
            filters=[EndOfSeriesCutFilter(2 * metadata["prediction_length"])],
            source=test_dir / "data.json",
        )

    @property
    def _link(self) -> str:
        raise NotImplementedError

    def _extract_data(
        self, path: Path
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        raise NotImplementedError


# -------------------------------------------------------------------------------------------------


def _get_frequency(monash_frequency: str) -> str:
    parts = monash_frequency.split("_")
    assert len(parts) <= 2, "invalid frequency (too many underscores)"
    if len(parts) == 1:
        return _get_base(parts[0])
    return _get_multiple(parts[0]) + _get_base(parts[1])


def _get_base(part: str) -> str:
    if part.lower().startswith("m"):
        if part.lower() == "monthly":
            return "M"
        if part.lower() == "minutely":
            return "min"
        raise ValueError(f"invalid frequency base {part}")
    return part.upper()[0]


def _get_multiple(part: str) -> str:
    if part.isnumeric():
        return part
    if part == "half":
        return "0.5"
    raise ValueError(f"invalid multiple string {part}")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This example shows how to fit a model and evaluate its predictions.
"""

# third party imports
import pandas as pd
import pytest

# first party imports
from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.rolling_dataset import (
    StepStrategy,
    generate_rolling_dataset,
)
from gluonts.evaluation import Evaluator
from gluonts.evaluation.backtest import make_evaluation_predictions
from gluonts.model.deepar import DeepAREstimator
from gluonts.mx.trainer import Trainer


def create_dynamic_dataset(
    start: str, length: int, num_dynamic: int
) -> ListDataset:
    """Create a ListDataset with dynamic values equal to the target."""
    return ListDataset(
        [
            {
                FieldName.TARGET: list(range(length)),
                FieldName.START: pd.Timestamp(start),
                FieldName.FEAT_DYNAMIC_REAL: [list(range(length))]
                * num_dynamic,
                FieldName.FEAT_DYNAMIC_CAT: [list(range(length))]
                * num_dynamic,
            }
        ],
        freq="D",
    )


@pytest.mark.parametrize(
    "train_length, test_length, prediction_length, target_start, rolling_start, num_dynamic_feat",
    [
        (10, 15, 2, "01-01-2019", "01-13-2019", 1),
        (10, 15, 2, "01-01-2019", "01-11-2019", 2),
    ],
)
def test_dynamic_integration(
    train_length: int,
    test_length: int,
    prediction_length: int,
    target_start: str,
    rolling_start: str,
    num_dynamic_feat: int,
):
    """
    Trains an estimator on a rolled dataset with dynamic features.
    Tests https://github.com/awslabs/gluonts/issues/1390
    """
    train_ds = create_dynamic_dataset(
        target_start, train_length, num_dynamic_feat
    )
    rolled_ds = generate_rolling_dataset(
        dataset=create_dynamic_dataset(
            target_start, test_length, num_dynamic_feat
        ),
        strategy=StepStrategy(prediction_length=prediction_length),
        start_time=pd.Timestamp(rolling_start),
    )
    estimator = DeepAREstimator(
        freq="D",
        prediction_length=prediction_length,
        context_length=2 * prediction_length,
        use_feat_dynamic_real=True,
        trainer=Trainer(epochs=1),
    )
    predictor = estimator.train(training_data=train_ds)
    forecast_it, ts_it = make_evaluation_predictions(
        rolled_ds, predictor=predictor, num_samples=100
    )
    training_agg_metrics, _ = Evaluator(num_workers=0)(ts_it, forecast_it)
    # it should have failed by this point if the dynamic features were wrong
    assert training_agg_metrics



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from ._base import Recommender
from ._factory import (
    create_ensemble_recommender,
    create_recommender,
    ENSEMBLE_RECOMMENDER_REGISTRY,
    RECOMMENDER_REGISTRY,
)
from .greedy import GreedyRecommender
from .optimal import OptimalRecommender
from .pareto import ParetoRecommender

__all__ = [
    "ENSEMBLE_RECOMMENDER_REGISTRY",
    "GreedyRecommender",
    "OptimalRecommender",
    "ParetoRecommender",
    "RECOMMENDER_REGISTRY",
    "Recommender",
    "create_ensemble_recommender",
    "create_recommender",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
from typing import Generic, List, Optional, TypeVar

T = TypeVar("T")


class CandidateGenerator(ABC, Generic[T]):
    """
    A candidate generator provides candidate model configurations to the
    recommender.
    """

    @abstractmethod
    def fit(self, configs: List[T]) -> None:
        """
        Fits the candidate generator on a list of model configurations seen
        during training.

        Args:
            configs: The model configurations seen during training.
        """

    @abstractmethod
    def generate(self, candidates: Optional[List[T]] = None) -> List[T]:
        """
        Generates a list of possible model configurations according to the
        strategy defined by the class.

        Args:
            candidates: If provided, every model configuration returned must be a member of this
                set.

        Returns:
            The generated model configurations.
        """



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass
from typing import List, Optional, Tuple
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tsbench.recommender.utils import pareto_efficiency_mask


@dataclass
class Dimension:
    """
    A dimension describes a combination of metric name and its display name.
    """

    metric: str
    displayName: str


@dataclass
class DataEntry:
    """
    A data entry describes a set of datapoints to be displayed and how they
    should be displayed.
    """

    data: pd.DataFrame
    color: str
    marker: str
    size: int
    edgecolor: Optional[str] = None
    edgewidth: Optional[int] = None
    name: Optional[str] = None
    alpha: float = 1


def plot_scatter_matrix(  # pylint: disable=too-many-statements
    data_entries: List[DataEntry],
    dimensions: List[Dimension],
    title: Optional[str] = None,
    savefig: Optional[str] = None,
    position_legend: bool = True,
    size: Tuple[int, int] = (4, 3),
    tick_fontsize: Optional[int] = None,
    axis_fontsize: Optional[int] = None,
    legend_fontsize: Optional[int] = None,
    legend_ncol: int = 1,
    ylog: bool = True,
    xmul: float = 1,
    plot_pareto_front: bool = False,
) -> None:
    """
    Creates a set of plots that visualize the data entries along the provided
    dimensions.
    """
    n = len(dimensions)
    fig, axes = plt.subplots(
        n - 1, n - 1, figsize=((n - 1) * size[0], (n - 1) * size[1]), dpi=150
    )
    if title is not None:
        fig.suptitle(title)

    for i in range(n - 1):
        for j in range(n - 1):
            ax = axes[i, j] if n > 2 else axes
            if j > i:
                ax.axis("off")
                continue

            xs = []
            ys = []
            for entry in data_entries:
                ax.scatter(
                    entry.data[dimensions[j].metric] * xmul,
                    entry.data[dimensions[i + 1].metric],
                    marker=entry.marker,  # type: ignore
                    s=entry.size,
                    color=entry.color,
                    edgecolors=entry.edgecolor,
                    linewidths=entry.edgewidth,
                    label=entry.name,
                    alpha=entry.alpha,
                )
                xs.extend((entry.data[dimensions[j].metric] * xmul).to_list())
                ys.extend(entry.data[dimensions[i + 1].metric].to_list())

            ax.set_xscale("log")
            if ylog:
                ax.set_yscale("log")

            if plot_pareto_front:
                ylim = ax.get_ylim()
                xlim = ax.get_xlim()
                xs = np.array(xs)
                ys = np.array(ys)
                all_points = np.stack([xs, ys], axis=1)
                mask = pareto_efficiency_mask(all_points)
                pareto_front = np.array(
                    sorted(all_points[mask], key=lambda r: -r[1])
                )
                pareto_front_x = pareto_front[:, 0].tolist()
                pareto_front_y = pareto_front[:, 1].tolist()
                ax.step(
                    [pareto_front_x[0]] + pareto_front_x + [1e10],
                    [1e10] + pareto_front_y + [pareto_front_y[-1]],
                    where="post",
                    color="black",
                    linewidth=1.5,
                    label="Pareto Front" if i == 0 else None,
                    zorder=-1000,
                )
                ax.set_ylim(ylim)
                ax.set_xlim(xlim)

            if j == 0:
                ax.set_ylabel(
                    dimensions[i + 1].displayName, fontsize=axis_fontsize
                )
            if tick_fontsize is not None:
                for tick in ax.yaxis.get_major_ticks():
                    tick.label.set_fontsize(tick_fontsize)

            if i == n - 2:
                ax.set_xlabel(
                    dimensions[j].displayName, fontsize=axis_fontsize
                )
            if tick_fontsize is not None:
                for tick in ax.xaxis.get_major_ticks():
                    tick.label.set_fontsize(tick_fontsize)

            handles, labels = ax.get_legend_handles_labels()

    legend_anchor = ax if n == 2 else fig  # type: ignore
    if position_legend:
        legend_anchor.legend(
            handles,  # type: ignore
            labels,  # type: ignore
            loc="upper right",
            bbox_to_anchor=(0, 0, 0.9, 0.9),
            fontsize=legend_fontsize,
            ncol=legend_ncol,
        )
    else:
        legend_anchor.legend(
            handles,  # type: ignore
            labels,  # type: ignore
            fontsize=legend_fontsize,
            ncol=legend_ncol,
        )
    plt.tight_layout()
    if savefig is not None:
        fig.patch.set_facecolor("white")
        plt.savefig(savefig)
    plt.show()
